

class CustomerService:
    api_url = "customers"

    def __init__(self, servicefusion):
        super().__init__(servicefusion)
    
    def get_customer_by_id(self, customer_id):
        params = {
            'expand': 'contacts.emails, custom_fields, contacts.phones, locations'
        }
        return self._get(f'{self.api_url}/{customer_id}', **params)

    def get_customer_by_name(self, firstname, lastname):
        params = {
            'filters[name]': f"{firstname} {lastname}",
            'fields': 'id',
            'expand': 'contacts.emails, contacts.phones'
        }
        return self._get(f'{self.api_url}', **params)

    def get_customer_by_email(self, customer_email):
        params = {
            'filters[email]': customer_email,
            'fields': 'id'
        }
        return self._get(f'{self.api_url}', **params)



import json
from json import JSONDecodeError
from pathlib import Path

from servicefusion.ServicefusionToken import ServicefusionToken
from servicefusion.storages import BaseStorage


class InMemoryStorage(BaseStorage):
    tokens: dict = {}

    def __init__(self):
        super().__init__()

    def get_token(self, app) -> ServicefusionToken:
        try:
            if app in self.tokens:
                return self.tokens[app]
        except Exception as e:
            pass
        return ServicefusionToken()

    def save_token(self, app: str, token: ServicefusionToken) -> bool:
        try:
            self.tokens[app] = token
            return True
        except Exception as e:
            pass
        return False



import logging
import os
from pathlib import Path
from logging.handlers import RotatingFileHandler
from flask import Blueprint
from servicefusion.REST.REST_V1 import REST_V1
from servicefusion.ServicefusionToken import ServicefusionToken
from .storages.JSONStorage import JSONStorage

from authlib.integrations.requests_client import OAuth2Session


class Servicefusion(OAuth2Session):
    def __init__(self, app):
        app_base_dir = app.root_path
        self.token_endpoint = "https://api.servicefusion.com/oauth/access_token"
        self.auth_endpoint = "https://api.servicefusion.com/oauth/authorize"

        self.logger = self.setup_logger(app_base_dir)
        self.token_updater = self.save_token
        self.token_handler = JSONStorage(Path(os.path.join(app_base_dir, "tokens/tokens.json")))

        OAuth2Session.__init__(self, client_id=app.config.get('SERVICEFUSION_CLIENT_ID'),
                               client_secret=app.config.get('SERVICEFUSION_CLIENT_SECRET'),
                               token_endpoint_auth_method="client_secret_basic", update_token=self.save_token)
        self.token = self.get_token()
        app.register_blueprint(self.get_blueprint())

    def save_token(self, token, **kwargs):
        sf_token = ServicefusionToken(**token)
        self.token = token
        self.token_handler.save_token('Servicefusion', sf_token)

    def get_token(self):
        token = self.token_handler.get_token('Servicefusion')
        if not token.exists:
            token = self.fetch_token(self.token_endpoint, grant_type='client_credentials')
            self.save_token(token)
            self.logger.info("Token Created")
            return token
        elif token.is_expired:
            self.refresh_token(url=self.token_endpoint, refresh_token=token.refresh_token)
        return token

    # def get_authorization_url(self):
    #     return self.create_authorization_url(self.auth_endpoint)
    #
    # def handle_callback(self, callback_request, code):
    #     token = self.fetch_token(url=self.token_endpoint, authorization_response=callback_request.url, scope='full',
    #                              client_id=self.client_id,
    #                              redirect_uri=self.redirect_uri,
    #                              client_secret=self.client_secret, code=code,
    #                              grant_type='authorization_code')
    #     self.auth_handler.store_token(token)

    def get_blueprint(self):
        servicefusion_blueprint = Blueprint('servicefusion_blueprint', __name__, cli_group="servicefusion")

        @servicefusion_blueprint.cli.command()
        def refresh_auth():
            token = self.token_handler.get_token('Servicefusion')
            self.refresh_token(url=self.token_endpoint, refresh_token=token.refresh_token)
            # self.

        return servicefusion_blueprint

    @staticmethod
    def setup_logger(base_dir):
        log_dir = (os.path.join(base_dir, "logs"))
        log_formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
        log_file_handler = RotatingFileHandler(log_dir + f'/servicefusion.log', maxBytes=100000, backupCount=5)
        log_file_handler.setFormatter(log_formatter)
        if not os.path.exists(log_dir):
            try:
                os.makedirs(log_dir)
            except Exception as e:
                print(e)
        logger = logging.getLogger("servicefusion")
        logger.addHandler(log_file_handler)
        logger.setLevel(logging.DEBUG)
        return logger




from datetime import datetime, timedelta
from pytz import timezone
from servicefusion.JSONHandler import JSONHandler


class JobService:
    api_url = "jobs"
    json_handler = JSONHandler()

    def __init__(self, servicefusion):
        super().__init__(servicefusion)

    @staticmethod
    def get_datetime():
        mst = timezone('MST')
        return datetime.now(mst)

    @staticmethod
    def get_rfc_date(date=None):
        if date is None:
            # return rfc3339.rfc3339(datetime.now(), use_system_timezone=False)
            return datetime.now().strftime("%Y-%m-%dT%H:%M:%S-00:00")
        else:
            # return rfc3339.rfc3339(date, use_system_timezone=False)
            return date.strftime("%Y-%m-%dT%H:%M:%S-00:00")

    def get_job_by_id(self, job_id):
        params = {
            'expand': 'equipment, equipment.custom_fields, products, services, other_charges, labor_charges, expenses, invoices'
        }
        response = self._get(f'{self.api_url}/{job_id}', **params)
        return response

    def get_jobs(self):
        filter_date = self.get_rfc_date(date=self.get_datetime() - timedelta(hours=1))
        current_page = 1
        complete = False
        results = []

        while not complete:
            params = {
                'page': current_page,
                'per-page': '50',
                'filters[updated_date][gte]': filter_date,
                'filters[status]': 'Job Closed, Completed, Partially Completed, Scheduled, Started, Waiting - Acrylic, '
                                   'Waiting - Circ Pump, Waiting - Wifi Card, Waiting for parts, Unscheduled, Delayed, '
                                   'On The Way, Cancelled, Invoiced',
                'expand': 'notes, contact',
            }
            response = self._get('jobs', **params)
            if self.json_handler.extract_value(response, 'totalCount') > 0:
                if 'items' in response:
                    results += response.get("items")
                    if response.get("currentPage") == response.get("pageCount"):
                        complete = True
                    else:
                        current_page = current_page + 1
                        complete = False
            else:
                return None
        return results



import json
from json import JSONDecodeError
from pathlib import Path

from servicefusion.ServicefusionToken import ServicefusionToken
from servicefusion.storages import BaseStorage


class JSONStorage(BaseStorage):
    def __init__(self, storage_path: Path = None):
        self.storage_path = storage_path
        super().__init__()

    def load_tokens(self) -> dict:
        tokens = {}
        try:
            with open(self.storage_path, "r") as f:
                tokens = json.load(f)
        except FileNotFoundError as e:
            with open(self.storage_path, "w") as f:
                json.dump(tokens, f, indent=4, ensure_ascii=True)
        except JSONDecodeError as e:
            with open(self.storage_path, "w") as f:
                json.dump(tokens, f, indent=4, ensure_ascii=True)
        return tokens

    def get_token(self, app) -> ServicefusionToken:
        tokens = self.load_tokens()
        try:
            if app in tokens:
                return ServicefusionToken(**tokens[app])
        except Exception as e:
            pass
        return ServicefusionToken()

    def save_token(self, app: str, token: ServicefusionToken) -> bool:
        try:
            tokens = self.load_tokens()
            tokens[app] = token.__dict__
            with open(self.storage_path, "w") as f:
                json.dump(tokens, f, indent=4, ensure_ascii=True)
            return True
        except Exception as e:
            pass
        return False

    def list_tokens_by_name(self):
        tokens = self.load_tokens()
        return list(tokens.keys())



from time import time


class ServicefusionToken:
    def __init__(self, **kwargs):
        self.access_token = kwargs.get("access_token", None)
        self.refresh_token = kwargs.get("refresh_token", None)
        self.expires_in = kwargs.get("expires_in", 0)
        self.expires_at = kwargs.get("expires_at")
        self.scope = kwargs.get("scope", None)

    @property
    def is_expired(self):
        return self.expires_at < int(time())

    @property
    def exists(self):
        return self.access_token is not None



import json
from datetime import datetime, date
from enum import Enum

import requests

from servicefusion.Servicefusion import ServicefusionToken, Servicefusion
from servicefusion.REST.models import BaseDataModel
from servicefusion.exceptions import ServicefusionUnauthorizedException, ServicefusionForbiddenException, ServicefusionTokenExpiredException


class RecordNotFoundError(Exception):
    pass


class BaseService:
    api_version = "v1"
    base_api_url = f"https://api.servicefusion.com/{api_version}/"
    api_url: str = None
    _service = None
    client = None
    Servicefusion: Servicefusion

    def __init__(self, Servicefusion):
        self.Servicefusion = Servicefusion

    @property
    def token(self) -> ServicefusionToken:
        return self.Servicefusion.token

    def _get(self, path: str, params: dict = None):
        return self.restful_request("GET", path, params=params)

    def _post(self, path: str, data: dict = None, params: dict = None):
        return self.restful_request("POST", path, data=data, params=params)

    def _delete(self, path: str, data: dict = None, params: dict = None):
        return self.restful_request("DELETE", path, data=data, params=params)

    def _put(self, path: str, data: dict = None, params: dict = None):
        return self.restful_request("PUT", path, data=data, params=params)

    def _patch(self, path: str, data: dict = None, params: dict = None):
        return self.restful_request("PATCH", path, data=data, params=params)

    def clean_data(self, value):
        if isinstance(value, list):
            for i, v in enumerate(value):
                value[i] = self.clean_data(v)
        if isinstance(value, datetime):
            value = f"{value.strftime('%Y-%m-%d')}T{value.strftime('%H:%M:%S.%f')}Z"
        if isinstance(value, date):
            value = f"{value.strftime('%Y-%m-%d')}"
        if isinstance(value, BaseDataModel):
            value = value.data
        if isinstance(value, Enum):
            value = value.value
        return value

    def serialize(self, data):
        # clean up data dict
        if "self" in data:
            del data['self']
        data = {k: v for k, v in data.items() if v}
        for k, v in data.items():
            data[k] = self.clean_data(v)
        return data

    def restful_request(self, method: str, path: str, data: dict = None, params: dict = None):
        if data is None:
            data = dict()
        if params is None:
            params = dict()
        params["access_token"] = self.token.access_token
        headers = {
            'Content-Type': 'application/json'
        }

        # clean up data
        data = self.serialize(data)
        params = self.serialize(params)

        url = f"{self.base_api_url}/{path}"
        if method.lower() in ['get', 'delete']:
            response = requests.request(method, url, headers=headers, params=params)
        else:
            response = requests.request(method, url, headers=headers, data=json.dumps(data), params=params)

        # TODO: Implement rate limiting here. response.headers contains the Apigee rate limiting info.
        if response.status_code == 204:  # No Content
            return True
        response_json = response.json()
        fault = response_json.get('fault', {}).get('faultstring', 'Error')
        if response.status_code == 401:  # Unauthorized
            raise ServicefusionUnauthorizedException(fault)
        if response.status_code == 403:  # Forbidden
            raise ServicefusionForbiddenException(fault)
        if response.status_code == 404:  # Not Found
            raise RecordNotFoundError(fault)

        return response_json

    @property
    def service(self):
        return self._service if self._service else self.__class__.__name__



import json


class JSONHandler:
    @staticmethod
    def get_from_storage(json_file):
        try:
            with open(json_file, "r", encoding="utf-8") as f:
                data = json.load(f)
                return data
        except FileNotFoundError:
            data = {}
            with open(json_file, "w", encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False, indent=4)
            return None

    @staticmethod
    def write_to_storage(json_file, data):
        with open(json_file, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=4)

    @staticmethod
    def extract_value(obj, key, return_empty_string=False):
        """Recursively fetch values from nested JSON."""
        arr = []

        def extract(obj, arr, key):
            """Recursively search for values of key in JSON tree."""
            if isinstance(obj, dict):
                for k, v in obj.items():
                    if isinstance(v, (dict, list)):
                        extract(v, arr, key)
                    elif k == key:
                        arr.append(v)
            elif isinstance(obj, list):
                for item in obj:
                    extract(item, arr, key)
            return arr

        values = extract(obj, arr, key)
        if values:
            return values[0]
        else:
            if return_empty_string:
                return ''
            else:
                return None

    @staticmethod
    def extract_values(obj, key):
        """Recursively fetch values from nested JSON."""
        arr = []

        def extract(obj, arr, key):
            """Recursively search for values of key in JSON tree."""
            if isinstance(obj, dict):
                for k, v in obj.items():
                    if isinstance(v, (dict, list)):
                        extract(v, arr, key)
                    elif k == key:
                        arr.append(v)
            elif isinstance(obj, list):
                for item in obj:
                    extract(item, arr, key)
            return arr

        values = extract(obj, arr, key)
        return values





from servicefusion.REST.V1 import ( CustomerService, JobService )

class REST_V1:
    def __init__(self, servicefusion):
        self.servicefusion = servicefusion
        self.CustomerService = CustomerService(servicefusion)
        self.JobService = JobService(servicefusion)




from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()


setup(
    name='flask_servicefusion',
    version='0.6.0',
    description='A servicefusion Rest Client For Flask applications',
    author='will sexton',
    author_email='will@theapiguys.com',
    url='https://bitbucket.org/will_sexton/flask_servicefusion/src/master/',
    readme="README.md",
    license='BSD 2-clause',
    packages=find_packages(),
    install_requires=['Flask',
                      'AuthLib',
                      'python-dateutil',
                      'requests'
                      ],
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3',
    ],
    long_description=long_description,
    long_description_content_type="text/markdown"

)



import json
from json import JSONDecodeError

from servicefusion import ServicefusionToken


class BaseStorage:
    app: str = None

    def __init__(self):
        pass

    def get_token(self, app) -> ServicefusionToken:
        raise NotImplementedError()

    def save_token(self, app: str, token: ServicefusionToken) -> bool:
        raise NotImplementedError()



class ServicefusionException(Exception):
    """Base exception for all library errors."""
    pass


class ServicefusionAuthError(ServicefusionException):
    """Something is wrong with authentication."""
    pass


class ServicefusionTokenError(ServicefusionAuthError):
    """Something is wrong with the tokens."""
    pass


class ServicefusionDataError(ServicefusionException, ValueError):
    """Something is wrong with the data."""
    pass


class ServicefusionConnectionError(ServicefusionException, ConnectionError):
    """Something is wrong with the connection."""
    pass

class ServicefusionException(Exception):
    pass


class ServicefusionForbiddenException(ServicefusionException):
    pass


class ServicefusionUnauthorizedException(ServicefusionException):
    pass


class ServicefusionTokenExpiredException(ServicefusionException):
    pass